import pytest
import requests
from typing import List
from pydantic import parse_obj_as
from bkw_python import BkwApi
from bkw_python.data_models import Case

def test_search_cases_single(requests_mock):
    requests_mock.get("https://api.bk.watch/api/2022-08-01?from=1&to=1&OPERATION=SearchCases&PROTOCOL=JSON&username=test&password=password", json={'results': {'items': [{'id': 'case-2f7ceff2', 'district': 'CAN', 'division': 'Oakland', 'caseNumber': '4:11-ap-4138', 'title': 'Silmi v. Fasail', 'dateFiled': '2011-04-29', 'dateTerminated': '2011-11-08', 'closed': True, 'voluntary': False}], 'offset': 1, 'total': 6369116}, 'status': 0, 'message': 'Success'})
    testing_session = BkwApi(username='test', password='password')
    cases = testing_session.search_cases(start=1, end=1)
    assert len(cases) == 1
    assert cases[0].caseNumber == "4:11-ap-4138"
    assert cases[0].id == "case-2f7ceff2"
    assert cases[0].district == "CAN"
    assert cases[0].dateDismissed == None
    assert cases[0].dateTerminated.strftime("%Y-%m-%d") == "2011-11-08"

def test_search_cases_multiple(requests_mock):
    requests_mock.get(
        "https://api.bk.watch/api/2022-08-01?username=test&password=password&PROTOCOL=JSON&OPERATION=SearchCases&from=1&to=10", 
        json={'results': {'items': [{'id': 'case-2f7ceff2', 'district': 'CAN', 'division': 'Oakland', 'caseNumber': '4:11-ap-4138', 'title': 'Silmi v. Fasail', 'dateFiled': '2011-04-29', 'dateTerminated': '2011-11-08', 'closed': True, 'voluntary': False}, {'id': 'case-0a482b11', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:10-ap-107', 'title': 'Warren, Trustee v. MetLife Home Loans, a Division of MetLife Bank, N.', 'dateFiled': '2010-05-12', 'dateTerminated': '2011-11-22', 'closed': True, 'voluntary': False}, {'id': 'case-9bb90769', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-204', 'title': 'Hicks v. Spring Leaf Financial', 'dateFiled': '2011-06-27', 'dateTerminated': '2011-09-09', 'closed': True, 'voluntary': False}, {'id': 'case-d0dd543d', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-274', 'title': 'Duren et al v. Coltate Capital, LLC', 'dateFiled': '2011-07-25', 'dateTerminated': '2011-10-20', 'closed': True, 'voluntary': False}, {'id': 'case-6eba400b', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-178', 'title': 'Sorrell et al v. Household Realty Corporation', 'dateFiled': '2011-06-01', 'dateTerminated': '2011-08-12', 'closed': True, 'voluntary': False}, {'id': 'case-8324e6c0', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-190', 'title': 'Melvin, II et al v. Beneficial', 'dateFiled': '2011-06-16', 'dateTerminated': '2011-08-12', 'closed': True, 'voluntary': False}, {'id': 'case-9bb90959', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-186', 'title': 'Mason et al v. Wells Fargo Bank, N.A.', 'dateFiled': '2011-06-13', 'dateTerminated': '2011-08-17', 'closed': True, 'voluntary': True}, {'id': 'case-2878eb03', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-253', 'title': 'Nelson v. American General Financial Services', 'dateFiled': '2011-06-30', 'dateTerminated': '2011-09-09', 'closed': True, 'voluntary': False}, {'id': 'case-2f7ce3a2', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:10-ap-309', 'title': 'Baldwin v. Fidelity National Title Insurance Company of New Y', 'dateFiled': '2010-12-08', 'dateTerminated': '2011-08-02', 'voluntary': True}, {'id': 'case-524f5fb7', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-163', 'title': 'Hart et al v. Green Tree Servicing, LLC', 'dateFiled': '2011-05-10', 'dateTerminated': '2011-08-03', 'closed': True, 'voluntary': False}], 'offset': 1, 'total': 6383131}, 'status': 0, 'message': 'Success'}
    )
    testing_session = BkwApi(username='test', password='password')
    cases = testing_session.search_cases(start=1, end=10)
    assert len(cases) == 10
    assert cases[0].id == 'case-2f7ceff2'
    assert cases[0].caseNumber == '4:11-ap-4138'
    assert cases[9].id == 'case-524f5fb7'
    assert cases[9].title == 'Hart et al v. Green Tree Servicing, LLC'

def test_search_cases_no_results(requests_mock):
    requests_mock.get(
        "https://api.bk.watch/api/2022-08-01?username=test&password=password&PROTOCOL=JSON&OPERATION=SearchCases&filters%5B%5D=chapter%2Cin%2C7&filters%5B%5D=district%2Cin%2CAK&filters%5B%5D=dateFiled%2Cge%2C2023-02-26&from=1&to=10",
        json={'results': {'total': 0}, 'status': 0, 'message': 'Success'}
    )
    testing_session = BkwApi(username='test', password='password')
    cases = testing_session.search_cases(start=1, end=10, districts=['AK'], chapters=[7], event="dateFiled", eventAfter="2023-02-26")
    assert len(cases) == 0

def test_search_cases_output_parties(requests_mock):
    requests_mock.get(
        "https://api.bk.watch/api/2022-08-01?username=test&password=password&PROTOCOL=JSON&OPERATION=SearchCases&outputFormat=parties&from=1&to=10",
        json={'results': {'items': [{'id': 'case-2f7ceff2', 'district': 'CAN', 'division': 'Oakland', 'caseNumber': '4:11-ap-4138', 'title': 'Silmi v. Fasail', 'dateFiled': '2011-04-29', 'dateTerminated': '2011-11-08', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-34abc01e', 'role': 'JUDGE', 'name': 'Edward D. Jellen', 'address': {'firstName': 'Edward', 'middleName': 'D.', 'lastName': 'Jellen'}}]}, {'id': 'case-0a482b11', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:10-ap-107', 'title': 'Warren, Trustee v. MetLife Home Loans, a Division of MetLife Bank, N.', 'dateFiled': '2010-05-12', 'dateTerminated': '2011-11-22', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-d0dd549d', 'role': 'JUDGE', 'name': 'Stephani W. Humrickhouse', 'address': {'firstName': 'Stephani', 'middleName': 'W.', 'lastName': 'Humrickhouse'}}]}, {'id': 'case-9bb90769', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-204', 'title': 'Hicks v. Spring Leaf Financial', 'dateFiled': '2011-06-27', 'dateTerminated': '2011-09-09', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-67af3b85', 'role': 'JUDGE', 'name': 'Stephani W. Humrickhouse', 'address': {'firstName': 'Stephani', 'middleName': 'W.', 'lastName': 'Humrickhouse'}}]}, {'id': 'case-d0dd543d', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-274', 'title': 'Duren et al v. Coltate Capital, LLC', 'dateFiled': '2011-07-25', 'dateTerminated': '2011-10-20', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-9bb907a9', 'role': 'JUDGE', 'name': 'J. Rich Leonard', 'address': {'firstName': 'J.', 'middleName': 'Rich', 'lastName': 'Leonard'}}]}, {'id': 'case-6eba400b', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-178', 'title': 'Sorrell et al v. Household Realty Corporation', 'dateFiled': '2011-06-01', 'dateTerminated': '2011-08-12', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-524f5c57', 'role': 'JUDGE', 'name': 'Stephani W. Humrickhouse', 'address': {'firstName': 'Stephani', 'middleName': 'W.', 'lastName': 'Humrickhouse'}}]}, {'id': 'case-8324e6c0', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-190', 'title': 'Melvin, II et al v. Beneficial', 'dateFiled': '2011-06-16', 'dateTerminated': '2011-08-12', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-a9114d6c', 'role': 'JUDGE', 'name': 'Stephani W. Humrickhouse', 'address': {'firstName': 'Stephani', 'middleName': 'W.', 'lastName': 'Humrickhouse'}}]}, {'id': 'case-9bb90959', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-186', 'title': 'Mason et al v. Wells Fargo Bank, N.A.', 'dateFiled': '2011-06-13', 'dateTerminated': '2011-08-17', 'closed': True, 'voluntary': True, 'parties': [{'id': 'party-67af3575', 'role': 'JUDGE', 'name': 'J. Rich Leonard', 'address': {'firstName': 'J.', 'middleName': 'Rich', 'lastName': 'Leonard'}}]}, {'id': 'case-2878eb03', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-253', 'title': 'Nelson v. American General Financial Services', 'dateFiled': '2011-06-30', 'dateTerminated': '2011-09-09', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-2fd829ff', 'role': 'JUDGE', 'name': 'J. Rich Leonard', 'address': {'firstName': 'J.', 'middleName': 'Rich', 'lastName': 'Leonard'}}]}, {'id': 'case-2f7ce3a2', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:10-ap-309', 'title': 'Baldwin v. Fidelity National Title Insurance Company of New Y', 'dateFiled': '2010-12-08', 'dateTerminated': '2011-08-02', 'voluntary': True, 'parties': [{'id': 'party-6eba431b', 'role': 'JUDGE', 'name': 'J. Rich Leonard', 'address': {'firstName': 'J.', 'middleName': 'Rich', 'lastName': 'Leonard'}}]}, {'id': 'case-524f5fb7', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-163', 'title': 'Hart et al v. Green Tree Servicing, LLC', 'dateFiled': '2011-05-10', 'dateTerminated': '2011-08-03', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-9fe89718', 'role': 'JUDGE', 'name': 'Stephani W. Humrickhouse', 'address': {'firstName': 'Stephani', 'middleName': 'W.', 'lastName': 'Humrickhouse'}}]}], 'offset': 1, 'total': 6383137}, 'status': 0, 'message': 'Success'}
    )
    testing_session = BkwApi(username='test', password='password')
    cases = testing_session.search_cases(start=1, end=10, districts=['AK'], outputFormat="parties")
    assert len(cases) == 10
    assert cases[0].parties[0].role == 'JUDGE'
    assert cases[0].parties[0].name == 'Edward D. Jellen'
    assert cases[0].parties[0].address.lastName == 'Jellen'

def test_search_cases_output_attorneys(requests_mock):
    requests_mock.get(
        "https://api.bk.watch/api/2022-08-01?username=test&password=password&PROTOCOL=JSON&OPERATION=SearchCases&outputFormat=parties&from=1&to=10",
        json={'results': {'items': [{'id': 'case-2f7ceff2', 'district': 'CAN', 'division': 'Oakland', 'caseNumber': '4:11-ap-4138', 'title': 'Silmi v. Fasail', 'dateFiled': '2011-04-29', 'dateTerminated': '2011-11-08', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-34abc01e', 'role': 'JUDGE', 'name': 'Edward D. Jellen', 'address': {'firstName': 'Edward', 'middleName': 'D.', 'lastName': 'Jellen'}}]}, {'id': 'case-0a482b11', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:10-ap-107', 'title': 'Warren, Trustee v. MetLife Home Loans, a Division of MetLife Bank, N.', 'dateFiled': '2010-05-12', 'dateTerminated': '2011-11-22', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-d0dd549d', 'role': 'JUDGE', 'name': 'Stephani W. Humrickhouse', 'address': {'firstName': 'Stephani', 'middleName': 'W.', 'lastName': 'Humrickhouse'}}]}, {'id': 'case-9bb90769', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-204', 'title': 'Hicks v. Spring Leaf Financial', 'dateFiled': '2011-06-27', 'dateTerminated': '2011-09-09', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-67af3b85', 'role': 'JUDGE', 'name': 'Stephani W. Humrickhouse', 'address': {'firstName': 'Stephani', 'middleName': 'W.', 'lastName': 'Humrickhouse'}}]}, {'id': 'case-d0dd543d', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-274', 'title': 'Duren et al v. Coltate Capital, LLC', 'dateFiled': '2011-07-25', 'dateTerminated': '2011-10-20', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-9bb907a9', 'role': 'JUDGE', 'name': 'J. Rich Leonard', 'address': {'firstName': 'J.', 'middleName': 'Rich', 'lastName': 'Leonard'}}]}, {'id': 'case-6eba400b', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-178', 'title': 'Sorrell et al v. Household Realty Corporation', 'dateFiled': '2011-06-01', 'dateTerminated': '2011-08-12', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-524f5c57', 'role': 'JUDGE', 'name': 'Stephani W. Humrickhouse', 'address': {'firstName': 'Stephani', 'middleName': 'W.', 'lastName': 'Humrickhouse'}}]}, {'id': 'case-8324e6c0', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-190', 'title': 'Melvin, II et al v. Beneficial', 'dateFiled': '2011-06-16', 'dateTerminated': '2011-08-12', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-a9114d6c', 'role': 'JUDGE', 'name': 'Stephani W. Humrickhouse', 'address': {'firstName': 'Stephani', 'middleName': 'W.', 'lastName': 'Humrickhouse'}}]}, {'id': 'case-9bb90959', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-186', 'title': 'Mason et al v. Wells Fargo Bank, N.A.', 'dateFiled': '2011-06-13', 'dateTerminated': '2011-08-17', 'closed': True, 'voluntary': True, 'parties': [{'id': 'party-67af3575', 'role': 'JUDGE', 'name': 'J. Rich Leonard', 'address': {'firstName': 'J.', 'middleName': 'Rich', 'lastName': 'Leonard'}}]}, {'id': 'case-2878eb03', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-253', 'title': 'Nelson v. American General Financial Services', 'dateFiled': '2011-06-30', 'dateTerminated': '2011-09-09', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-2fd829ff', 'role': 'JUDGE', 'name': 'J. Rich Leonard', 'address': {'firstName': 'J.', 'middleName': 'Rich', 'lastName': 'Leonard'}}]}, {'id': 'case-2f7ce3a2', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:10-ap-309', 'title': 'Baldwin v. Fidelity National Title Insurance Company of New Y', 'dateFiled': '2010-12-08', 'dateTerminated': '2011-08-02', 'voluntary': True, 'parties': [{'id': 'party-6eba431b', 'role': 'JUDGE', 'name': 'J. Rich Leonard', 'address': {'firstName': 'J.', 'middleName': 'Rich', 'lastName': 'Leonard'}}]}, {'id': 'case-524f5fb7', 'district': 'NCE', 'division': 'Wilson', 'caseNumber': '8:11-ap-163', 'title': 'Hart et al v. Green Tree Servicing, LLC', 'dateFiled': '2011-05-10', 'dateTerminated': '2011-08-03', 'closed': True, 'voluntary': False, 'parties': [{'id': 'party-9fe89718', 'role': 'JUDGE', 'name': 'Stephani W. Humrickhouse', 'address': {'firstName': 'Stephani', 'middleName': 'W.', 'lastName': 'Humrickhouse'}}]}], 'offset': 1, 'total': 6383137}, 'status': 0, 'message': 'Success'}
    )
    testing_session = BkwApi(username='test', password='password')
    cases = testing_session.search_cases(start=1, end=10, districts=['AK'], outputFormat="parties")
    assert len(cases) == 10
    assert cases[0].parties[0].role == 'JUDGE'
    assert cases[0].parties[0].name == 'Edward D. Jellen'
    assert cases[0].parties[0].address.lastName == 'Jellen'

def test_search_cases_custom_filter_single_case(requests_mock):
    requests_mock.get(
        "https://api.bk.watch/api/2022-08-01?username=test&password=password&PROTOCOL=JSON&OPERATION=SearchCases&outputFormat=creditors&filters%5B%5D=id%2Cin%2Ccase-2f1baa62&from=1&to=10",
        json={'results': {'items': [{'id': 'case-2f1baa62', 'district': 'UT', 'division': 'Salt Lake City', 'county': 'Salt Lake', 'caseNumber': '2:22-bk-24237', 'title': 'Blake Lee Smith', 'chapter': 7, 'dateDischarged': '2023-02-01', 'dateEntered': '2022-10-27', 'dateFiled': '2022-10-27', 'closed': False, 'disposition': 'Standard Discharge', 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'BUSINESS'}], 'offset': 1, 'total': 1}, 'status': 0, 'message': 'Success'}
    )
    testing_session = BkwApi(username='test', password='password')
    cases = testing_session.search_cases(start=1, end=10, outputFormat='creditors', filters=['id,in,case-2f1baa62'])
    assert len(cases) == 1
    assert cases[0].title == 'Blake Lee Smith'
    assert cases[0].debtType == 'BUSINESS'

def test_search_cases_output_parties_single(requests_mock):
    requests_mock.get(
        "https://api.bk.watch/api/2022-08-01?username=test&password=password&PROTOCOL=JSON&OPERATION=SearchCases&filters%5B%5D=id%2Cin%2Ccase-2f1959d2&outputFormat=parties&from=1&to=1",
        json={'results': {'items': [{'id': 'case-2f1959d2', 'district': 'UT', 'division': 'St. George', 'county': 'Iron', 'caseNumber': '4:22-bk-24969', 'title': 'Shaunna Jeanne Manchester', 'chapter': 7, 'dateEntered': '2022-12-21', 'dateFiled': '2022-12-21', 'filingFeeStatus': 'Paid', 'assets': False, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER', 'parties': [{'id': 'party-50eed877', 'role': 'JUDGE', 'name': 'William T. Thurman', 'address': {'firstName': 'William', 'middleName': 'T.', 'lastName': 'Thurman'}}, {'id': 'party-a3b36d96', 'role': 'TRUSTEE', 'name': 'Jones', 'address': {'lastName': 'Jones'}}, {'id': 'party-650eb005', 'role': 'DEBTOR', 'name': 'Shaunna Jeanne Manchester', 'address': {'firstName': 'Shaunna', 'middleName': 'Jeanne', 'lastName': 'Manchester', 'street1': '76 N. 1050 W., Basement', 'city': 'Cedar City', 'state': 'UT', 'zip': '84720', 'ssn': 'xxx-xx-4939'}}, {'id': 'party-af16d924', 'role': 'US_TRUSTEE', 'name': 'United States Trustee', 'address': {'organization': 'United States Trustee', 'street1': 'Washington Federal Bank Bldg.', 'street2': '405 South Main Street', 'street3': 'Suite 300', 'city': 'Salt Lake City', 'state': 'UT', 'zip': '84111'}}]}], 'offset': 1, 'total': 1}, 'status': 0, 'message': 'Success'}
    )
    testing_session = BkwApi(username='test', password='password')
    cases = testing_session.search_cases(start=1, end=1, outputFormat='parties', filters=['id,in,case-2f1959d2'])
    assert len(cases) == 1
    assert len(cases[0].parties) == 4
    assert cases[0].parties[3].id == 'party-af16d924'
    assert cases[0].parties[3].role == 'US_TRUSTEE'

def test_search_cases_output_parties_multiple(requests_mock):
    requests_mock.get(
        "https://api.bk.watch/api/2022-08-01?username=test&password=password&PROTOCOL=JSON&OPERATION=SearchCases&filters%5B%5D=id%2Cin%2Ccase-2f1959d2%2Ccase-9bdc15c9&outputFormat=parties&from=1&to=2",
        json={'results': {'items': [{'id': 'case-2f1959d2', 'district': 'UT', 'division': 'St. George', 'county': 'Iron', 'caseNumber': '4:22-bk-24969', 'title': 'Shaunna Jeanne Manchester', 'chapter': 7, 'dateEntered': '2022-12-21', 'dateFiled': '2022-12-21', 'filingFeeStatus': 'Paid', 'assets': False, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER', 'parties': [{'id': 'party-50eed877', 'role': 'JUDGE', 'name': 'William T. Thurman', 'address': {'firstName': 'William', 'middleName': 'T.', 'lastName': 'Thurman'}}, {'id': 'party-a3b36d96', 'role': 'TRUSTEE', 'name': 'Jones', 'address': {'lastName': 'Jones'}}, {'id': 'party-650eb005', 'role': 'DEBTOR', 'name': 'Shaunna Jeanne Manchester', 'address': {'firstName': 'Shaunna', 'middleName': 'Jeanne', 'lastName': 'Manchester', 'street1': '76 N. 1050 W., Basement', 'city': 'Cedar City', 'state': 'UT', 'zip': '84720', 'ssn': 'xxx-xx-4939'}}, {'id': 'party-af16d924', 'role': 'US_TRUSTEE', 'name': 'United States Trustee', 'address': {'organization': 'United States Trustee', 'street1': 'Washington Federal Bank Bldg.', 'street2': '405 South Main Street', 'street3': 'Suite 300', 'city': 'Salt Lake City', 'state': 'UT', 'zip': '84111'}}]}, {'id': 'case-9bdc15c9', 'district': 'UT', 'division': 'Salt Lake City', 'county': 'Salt Lake', 'caseNumber': '2:23-bk-20183', 'title': 'Maria De Jesus Bacilio Guzman', 'chapter': 7, 'dateEntered': '2023-01-19', 'dateFiled': '2023-01-19', 'closed': False, 'filingFeeStatus': 'Installment', 'assets': False, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER', 'parties': [{'id': 'party-6c1e3f6b', 'role': 'JUDGE', 'name': 'Joel T. Marker', 'address': {'firstName': 'Joel', 'middleName': 'T.', 'lastName': 'Marker'}}, {'id': 'party-0b93543a', 'role': 'TRUSTEE', 'name': 'Michael F. Thomson', 'address': {'firstName': 'Michael', 'middleName': 'F.', 'lastName': 'Thomson', 'organization': 'Greenberg Traurig LLP', 'street1': '222 S MAIN ST Suite 1730', 'city': 'Salt Lake City', 'state': 'UT', 'zip': '84101', 'zip4': '2185', 'phone1': '801-478-6917', 'email': 'thomsonm@gtlaw.com'}}, {'id': 'party-991d7769', 'role': 'DEBTOR', 'name': 'Maria De Jesus Bacilio Guzman', 'address': {'firstName': 'Maria', 'middleName': 'De Jesus', 'lastName': 'Bacilio Guzman', 'street1': '3449 Maple Way', 'city': 'West Valley City', 'state': 'UT', 'zip': '84119', 'ssn': 'xxx-xx-0169'}}, {'id': 'party-9d4ceb98', 'role': 'US_TRUSTEE', 'name': 'United States Trustee', 'address': {'organization': 'United States Trustee', 'street1': 'Washington Federal Bank Bldg.', 'street2': '405 South Main Street', 'street3': 'Suite 300', 'city': 'Salt Lake City', 'state': 'UT', 'zip': '84111', 'email': 'USTPRegion19.SK.ECF@usdoj.gov'}}, {'id': 'party-650b4245', 'role': 'DEBTOR', 'name': 'Maria De Jesus Bacilio Guzman', 'address': {'firstName': 'Maria', 'middleName': 'De Jesus Bacilio', 'lastName': 'Guzman', 'street1': '3449 Maple Way', 'city': 'West Valley City', 'county': 'Salt Lake', 'state': 'UT', 'zip': '84119', 'ssn': 'xxx-xx-0169'}}]}], 'offset': 1, 'total': 2}, 'status': 0, 'message': 'Success'}
    )
    testing_session = BkwApi(username='test', password='password')
    cases = testing_session.search_cases(start=1, end=2, outputFormat='parties', filters=['id,in,case-2f1959d2,case-9bdc15c9'])
    assert len(cases) == 2
    assert len(cases[0].parties) == 4
    assert cases[0].parties[3].id == 'party-af16d924'
    assert cases[0].parties[3].role == 'US_TRUSTEE'
    assert cases[1].parties[3].id == 'party-9d4ceb98'
    assert cases[1].parties[2].id == 'party-991d7769'
    assert cases[1].parties[2].role == 'DEBTOR'

def test_search_cases_multiple_filters(requests_mock):
    requests_mock.get(
        "https://api.bk.watch/api/2022-08-01?filters%5B%5D=district%2Cin%2CTXN%2CTXE&filters%5B%5D=chapter%2Cin%2C13&filters%5B%5D=dateDismissed%2Cge%2C2023-02-01&from=1&to=50&OPERATION=SearchCases&PROTOCOL=JSON&username=test&password=password",
        json={'results': {'items': [{'id': 'case-091f805a', 'district': 'TXE', 'division': 'Sherman', 'county': 'Denton', 'caseNumber': '4:18-bk-40702', 'title': 'Tammie Toni Riggs', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2018-04-02', 'dateFiled': '2018-04-02', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-a932380c', 'district': 'TXE', 'division': 'Sherman', 'county': 'Dallas', 'caseNumber': '4:18-bk-41791', 'title': 'Christina Biggerstaff', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2018-08-13', 'dateFiled': '2018-08-13', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-a135b536', 'district': 'TXE', 'division': 'Beaumont', 'county': 'Jefferson', 'caseNumber': '1:18-bk-10455', 'title': 'Sharon Ann Byrd', 'chapter': 13, 'dateDismissed': '2023-02-22', 'dateEntered': '2018-11-05', 'dateFiled': '2018-11-05', 'datePlanConfirmed': '2019-07-30', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-5270a157', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:19-bk-40852', 'title': 'Sandra Elaine Branch', 'chapter': 13, 'dateDismissed': '2023-02-17', 'dateEntered': '2019-04-01', 'dateFiled': '2019-04-01', 'datePlanConfirmed': '2019-11-08', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-0a75bd91', 'district': 'TXE', 'division': 'Beaumont', 'county': 'Jasper', 'caseNumber': '1:19-bk-10216', 'title': 'Donald W. LeDeaux and Shirley D. LeDeaux', 'chapter': 13, 'dateDismissed': '2023-02-22', 'dateEntered': '2019-05-08', 'dateFiled': '2019-05-08', 'dateJointDebtorDismissed': '2023-02-22', 'datePlanConfirmed': '2019-11-18', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-2f4e17c2', 'district': 'TXE', 'division': 'Sherman', 'county': 'Denton', 'caseNumber': '4:19-bk-41348', 'title': 'Guadalupe R. Barelas', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2019-05-20', 'dateFiled': '2019-05-20', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-6e8ecfcb', 'district': 'TXE', 'division': 'Sherman', 'county': 'Dallas', 'caseNumber': '4:19-bk-42403', 'title': 'Telina D. Johnson', 'chapter': 13, 'dateDismissed': '2023-02-10', 'dateEntered': '2019-09-02', 'dateFiled': '2019-09-02', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-2fed43ef', 'district': 'TXE', 'division': 'Sherman', 'county': 'Dallas', 'caseNumber': '4:19-bk-42579', 'title': 'Crystal S. Johnson', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2019-09-20', 'dateFiled': '2019-09-20', 'datePlanConfirmed': '2019-12-20', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-0a028441', 'district': 'TXE', 'division': 'Beaumont', 'county': 'Jasper', 'caseNumber': '1:19-bk-10471', 'title': 'Corey James Dorman', 'chapter': 13, 'dateDismissed': '2023-02-17', 'dateEntered': '2019-10-03', 'dateFiled': '2019-10-03', 'closed': False, 'disposition': ' ', 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-d096961d', 'district': 'TXE', 'division': 'Sherman', 'county': 'Dallas', 'caseNumber': '4:19-bk-42841', 'title': 'Chandra Marisa Payton Williams and Patrick Lacharles Williams', 'chapter': 13, 'dateDismissed': '2023-02-10', 'dateEntered': '2019-10-18', 'dateFiled': '2019-10-18', 'dateJointDebtorDismissed': '2023-02-10', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-5207ba37', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:19-bk-42980', 'title': 'Lindan Lee Johnson', 'chapter': 13, 'dateDismissed': '2023-02-24', 'dateEntered': '2019-11-04', 'dateFiled': '2019-11-04', 'datePlanConfirmed': '2020-07-17', 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-adfe87a4', 'district': 'TXE', 'division': 'Beaumont', 'county': 'Jefferson', 'caseNumber': '1:19-bk-10537', 'title': 'Lesia M. Evans', 'chapter': 13, 'dateDismissed': '2023-02-24', 'dateEntered': '2019-11-21', 'dateFiled': '2019-11-21', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-836b48a0', 'district': 'TXE', 'division': 'Sherman', 'county': 'Tarrant', 'caseNumber': '4:20-bk-40012', 'title': 'Jackson Agnew-Stevens and Roshineke Deshontrice Agnew', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2020-01-02', 'dateFiled': '2020-01-02', 'dateJointDebtorDismissed': '2023-02-08', 'datePlanConfirmed': '2020-06-19', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-67e02b25', 'district': 'TXE', 'division': 'Sherman', 'county': 'Denton', 'caseNumber': '4:20-bk-40135', 'title': 'Adrian K Dove', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2020-01-13', 'dateFiled': '2020-01-13', 'closed': False, 'disposition': ' ', 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-adfb88f4', 'district': 'TXE', 'division': 'Texarkana', 'county': 'Bowie', 'caseNumber': '5:20-bk-50011', 'title': 'Cal Manning and Cindy Manning', 'chapter': 13, 'dateDismissed': '2023-02-10', 'dateEntered': '2020-01-17', 'dateFiled': '2020-01-17', 'dateJointDebtorDismissed': '2023-02-10', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-9bf5a2e9', 'district': 'TXE', 'division': 'Beaumont', 'county': 'Newton', 'caseNumber': '1:20-bk-10030', 'title': 'Roosevelt S Spikes, Sr', 'chapter': 13, 'dateDismissed': '2023-02-22', 'dateEntered': '2020-01-21', 'dateFiled': '2020-01-21', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-0971080a', 'district': 'TXE', 'division': 'Sherman', 'county': 'Dallas', 'caseNumber': '4:20-bk-40653', 'title': 'Oliver Benjamin Crowe and Pamela Arnetta Crowe', 'chapter': 13, 'dateDismissed': '2023-02-23', 'dateEntered': '2020-03-02', 'dateFiled': '2020-03-02', 'dateJointDebtorDismissed': '2023-02-23', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-0971083a', 'district': 'TXE', 'division': 'Sherman', 'county': 'Denton', 'caseNumber': '4:20-bk-40669', 'title': 'James R Trucksess', 'chapter': 13, 'dateDismissed': '2023-02-23', 'dateEntered': '2020-03-02', 'dateFiled': '2020-03-02', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-09713fea', 'district': 'TXE', 'division': 'Sherman', 'county': 'Dallas', 'caseNumber': '4:20-bk-40685', 'title': 'Brian L Compton and Shalena Compton', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2020-03-03', 'dateFiled': '2020-03-03', 'dateJointDebtorDismissed': '2023-02-08', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-83600f30', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:20-bk-40868', 'title': 'Daniel Allen Desandre and Stephanie C Desandre', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2020-03-27', 'dateFiled': '2020-03-27', 'dateJointDebtorDismissed': '2023-02-08', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-d099daad', 'district': 'TXE', 'division': 'Lufkin', 'county': 'Angelina', 'caseNumber': '9:20-bk-90097', 'title': 'Susan Smith Crawford', 'chapter': 13, 'dateDismissed': '2023-02-22', 'dateEntered': '2020-04-09', 'dateFiled': '2020-04-09', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-d09888ad', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:20-bk-41109', 'title': 'Robert Masterson, II', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2020-05-04', 'dateFiled': '2020-05-04', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-5215c067', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:20-bk-41336', 'title': 'Ivan F. Novotny', 'chapter': 13, 'dateDismissed': '2023-02-10', 'dateEntered': '2020-06-09', 'dateFiled': '2020-06-09', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-2f27c892', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:20-bk-41608', 'title': 'Linus Francis Walton', 'chapter': 13, 'dateDismissed': '2023-02-10', 'dateEntered': '2020-07-21', 'dateFiled': '2020-07-21', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-0a108261', 'district': 'TXE', 'division': 'Sherman', 'county': 'Dallas', 'caseNumber': '4:20-bk-41698', 'title': 'Sotero Munoz Martinez', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2020-08-03', 'dateFiled': '2020-08-03', 'datePlanConfirmed': '2020-12-04', 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-2821bca3', 'district': 'TXE', 'division': 'Sherman', 'county': 'Dallas', 'caseNumber': '4:20-bk-41932', 'title': 'Christine Walsh Shanks', 'chapter': 13, 'dateDismissed': '2023-02-22', 'dateEntered': '2020-09-11', 'dateFiled': '2020-09-11', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-52118d47', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:20-bk-42021', 'title': 'Maverick Dion Davis and Belinda Joyce Davis', 'chapter': 13, 'dateDismissed': '2023-02-22', 'dateEntered': '2020-09-25', 'dateFiled': '2020-09-25', 'dateJointDebtorDismissed': '2023-02-22', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-52119217', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:20-bk-42032', 'title': 'Ronald Edward Brown', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2020-09-28', 'dateFiled': '2020-09-28', 'datePlanConfirmed': '2021-02-05', 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-2f229572', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:20-bk-42082', 'title': 'Madeline Yolanda Delgado', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2020-10-05', 'dateFiled': '2020-10-05', 'datePlanConfirmed': '2020-12-18', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-34f59c7e', 'district': 'TXE', 'division': 'Tyler', 'county': 'Smith', 'caseNumber': '6:20-bk-60521', 'title': 'James Lee Carpenter', 'chapter': 13, 'dateDismissed': '2023-02-03', 'dateEntered': '2020-10-08', 'dateFiled': '2020-10-08', 'closed': False, 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-6ee51b7b', 'district': 'TXE', 'division': 'Beaumont', 'county': 'Jefferson', 'caseNumber': '1:20-bk-10449', 'title': 'David Jones', 'chapter': 13, 'dateDismissed': '2023-02-17', 'dateEntered': '2020-11-05', 'dateFiled': '2020-11-05', 'datePlanConfirmed': '2021-01-21', 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-2f211d62', 'district': 'TXE', 'division': 'Beaumont', 'county': 'San Jacinto', 'caseNumber': '1:20-bk-10506', 'title': 'Ronald D. Gilbert', 'chapter': 13, 'dateDismissed': '2023-02-22', 'dateEntered': '2020-12-17', 'dateFiled': '2020-12-17', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-2f8aedcf', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:21-bk-40055', 'title': 'Felicia Michelle Mitchell', 'chapter': 13, 'dateDismissed': '2023-02-22', 'dateEntered': '2021-01-13', 'dateFiled': '2021-01-13', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-6ee8cd0b', 'district': 'TXE', 'division': 'Sherman', 'county': 'Denton', 'caseNumber': '4:21-bk-40098', 'title': 'Stephanie Ellen Brackeen', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2021-01-21', 'dateFiled': '2021-01-21', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-09640b5a', 'district': 'TXE', 'division': 'Beaumont', 'county': 'Jasper', 'caseNumber': '1:21-bk-10072', 'title': 'Keelyn T. Stanley', 'chapter': 13, 'dateDismissed': '2023-02-17', 'dateEntered': '2021-03-11', 'dateFiled': '2021-03-11', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-282eb3a3', 'district': 'TXE', 'division': 'Sherman', 'county': 'Dallas', 'caseNumber': '4:21-bk-40712', 'title': 'Daniel J. Llana', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2021-05-12', 'dateFiled': '2021-05-12', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-2f8f7c4f', 'district': 'TXE', 'division': 'Sherman', 'county': 'Denton', 'caseNumber': '4:21-bk-40856', 'title': 'Michael Edward Oliver and Renee Carolyn Machado-Oliver', 'chapter': 13, 'previousChapter': 7, 'dateConverted': '2021-08-27', 'dateDismissed': '2023-02-10', 'dateEntered': '2021-06-10', 'dateFiled': '2021-06-10', 'dateJointDebtorDismissed': '2023-02-10', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-34c1452e', 'district': 'TXE', 'division': 'Texarkana', 'county': 'Bowie', 'caseNumber': '5:21-bk-50076', 'title': 'James M Thomas and Imogene A Thomas', 'chapter': 13, 'dateDismissed': '2023-02-10', 'dateEntered': '2021-08-28', 'dateFiled': '2021-08-28', 'dateJointDebtorDismissed': '2023-02-10', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-d0b6c29d', 'district': 'TXE', 'division': 'Sherman', 'county': 'Denton', 'caseNumber': '4:21-bk-41344', 'title': 'Donald Lee Baker, Jr.', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2021-09-22', 'dateFiled': '2021-09-22', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-28137b83', 'district': 'TXE', 'division': 'Beaumont', 'county': 'Jefferson', 'caseNumber': '1:21-bk-10305', 'title': 'Diane Jackson', 'chapter': 13, 'dateDismissed': '2023-02-17', 'dateEntered': '2021-09-23', 'dateFiled': '2021-09-23', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-2f17bb02', 'district': 'TXE', 'division': 'Texarkana', 'county': 'Bowie', 'caseNumber': '5:21-bk-50085', 'title': 'Steven Tilley and Tina Tilley', 'chapter': 13, 'dateDismissed': '2023-02-13', 'dateEntered': '2021-09-30', 'dateFiled': '2021-09-30', 'dateJointDebtorDismissed': '2023-02-13', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-67c40105', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:21-bk-41422', 'title': 'Shelia Frances Echols', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2021-10-06', 'dateFiled': '2021-10-06', 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-9f802cf8', 'district': 'TXE', 'division': 'Lufkin', 'county': 'Tyler', 'caseNumber': '9:21-bk-90200', 'title': 'Gary Adam Tousha and Amanda Rae Tousha', 'chapter': 13, 'dateDismissed': '2023-02-17', 'dateEntered': '2021-10-22', 'dateFiled': '2021-10-22', 'dateJointDebtorDismissed': '2023-02-17', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-a17a7756', 'district': 'TXE', 'division': 'Sherman', 'county': 'Denton', 'caseNumber': '4:21-bk-41526', 'title': 'Michelle Yvette Williams', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2021-10-27', 'dateFiled': '2021-10-27', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-2f1465d2', 'district': 'TXE', 'division': 'Sherman', 'county': 'Denton', 'caseNumber': '4:21-bk-41542', 'title': 'Elke Deneen Barber', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2021-10-29', 'dateFiled': '2021-10-29', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-a978958c', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:21-bk-41627', 'title': 'Roselaine Alcindor Cezaire', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2021-11-16', 'dateFiled': '2021-11-16', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-2f158902', 'district': 'TXE', 'division': 'Sherman', 'county': 'Denton', 'caseNumber': '4:21-bk-41719', 'title': 'Christopher Keith Raybon', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2021-12-02', 'dateFiled': '2021-12-02', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-0a217bf1', 'district': 'TXE', 'division': 'Sherman', 'county': 'Dallas', 'caseNumber': '4:21-bk-41739', 'title': 'Sharetta Spearman', 'chapter': 13, 'dateDismissed': '2023-02-06', 'dateEntered': '2021-12-06', 'dateFiled': '2021-12-06', 'datePlanConfirmed': '2022-03-03', 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-522648c7', 'district': 'TXE', 'division': 'Sherman', 'county': 'Collin', 'caseNumber': '4:21-bk-41790', 'title': 'Patrick Lasean Minnifield', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2021-12-15', 'dateFiled': '2021-12-15', 'filingFeeStatus': 'Paid', 'assets': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}, {'id': 'case-9f819158', 'district': 'TXE', 'division': 'Sherman', 'county': 'Dallas', 'caseNumber': '4:21-bk-41801', 'title': 'Stephen Shayne Booth', 'chapter': 13, 'dateDismissed': '2023-02-08', 'dateEntered': '2021-12-17', 'dateFiled': '2021-12-17', 'datePlanConfirmed': '2022-06-13', 'filingFeeStatus': 'Paid', 'assets': True, 'voluntary': True, 'debtorType': 'INDIVIDUAL', 'debtType': 'CONSUMER'}], 'offset': 1, 'total': 290}, 'status': 0, 'message': 'Success'}
    )
    testing_session = BkwApi(username='test', password='password')
    cases = testing_session.search_cases(districts=["TXN", "TXE"], chapters=[13], outputFormat="parties", event='dateDismissed', eventAfter="2023-02-01")
    assert len(cases) == 50
    assert cases[0].id == 'case-091f805a'